#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "bilateral.h"
#include "image_types.h"
#include "integral_histogram.h"

#define FILTER_JOB_SHARE 0.75
#define ENHANCE_JOB_SHARE 0.25

uint32_t bin_convert(uint32_t x,
                     uint32_t y,
                     const s3d_merge::Image8 *img,
                     void *data)
{
    int *channel = (int *)data;

    return uint32_t((img->get_pixel(x, y)[*channel])/BIN_SIZE);
}


// Filter context contains precalculated weights and mean values for various
// bins and bin/offset combinations.
typedef struct _filter_context
{
    uint32_t bin_map[256], offset_map[256];
    float bin_value[NUM_BINS];
    float offset_weights[NUM_BINS][BIN_SIZE];
} filter_context;

// Integrals of filter weights as a function of distance from centre.
static float linear_integral_fun(float threshold, float x)
{
    float integral = x - ((x * x) / (2.0 * threshold));
    return integral;
}

static float quadratic_integral_fun(float threshold, float x)
{
    float integral = x - ((x * x * x) / (3.0 * threshold * threshold));
    return integral;
}

// Initialisation of various lookup tables for bin average intensity values,
// bin weights for various initial offsets etc.
void initialise_filter_context(uint32_t threshold, 
                               bool quadratic, 
                               filter_context &ctx)
{
    float (*get_integral)(float, float)(NULL);
    float ft;
    ft = float(threshold)/256.0;

    if(quadratic)
    {
        get_integral = quadratic_integral_fun;
    }
    else
    {
        get_integral = linear_integral_fun;
    }

    for(uint32_t i=0;i<256;i++)
    {
        ctx.bin_map[i] = div(i,BIN_SIZE).quot;
        ctx.offset_map[i] = div(i,BIN_SIZE).rem;
    }

    for(uint32_t i=0;i<NUM_BINS;i++)
    {
        ctx.bin_value[i] = ((i * BIN_SIZE) + ((i+1) * BIN_SIZE))/2.0;

        for(uint32_t j=0;j<BIN_SIZE;j++)
        {
            uint32_t near_dist, far_dist;

            near_dist = j + (i * BIN_SIZE);
            far_dist = near_dist + BIN_SIZE;

            // Spot the edge of the filter kernel.
            if(far_dist <= threshold)
            {
                float near(near_dist), far(far_dist);
                near/= 256;
                far/= 256;
                ctx.offset_weights[i][j] = get_integral(ft, far) - get_integral(ft, near);
            }
            else if (near_dist <= threshold)
            {
                float near(near_dist), far(threshold);
                near/= 256;
                far/= 256;
                ctx.offset_weights[i][j] = get_integral(ft, far) - get_integral(ft, near);
            }
            else
            {
                ctx.offset_weights[i][j] = -1;
            }
        }
    }
}

void filter_tile(const s3d_merge::IntegralHistogram *hist,
                 const filter_context &ctx,
                 uint32_t radius,
                 uint32_t x_offset,
                 uint32_t y_offset,
                 uint32_t width,
                 uint32_t height,
                 uint32_t channel,
                 double min_progress,
                 double max_progress,
                 s3d_merge::Image8 *dest)
{
    uint32_t channels;

    channels = dest->get_channels();

    if((width + x_offset) > dest->get_width())
    {
        width = dest->get_width() - x_offset;
    }
    if((height + y_offset) > dest->get_height())
    {
        height = dest->get_height() - y_offset;
    }

    if(hist && dest)
    {
        for(uint32_t y=0;y<height;y++)
        {
            double progress;
            uint32_t ymax, row_index;
            uint8_t *row;

            row_index = channel;
            ymax = y + (radius * 2);// - 1;
            {
                uint32_t offset;
                offset  = (x_offset + ((y + y_offset) * dest->get_width()))*channels;
                row = dest->get_buffer() + offset;
            }

            for(uint32_t x=0;x<width;x++)
            {
                uint32_t bins[NUM_BINS];
                uint32_t xmax;

                uint32_t cur_val, cur_bin, offset, value;
                float total_value, total_weight;
                xmax = x + (radius * 2);

                // Get a histogram for the sub region.

                hist->GetHistogram(x, y, xmax, ymax, bins);

                cur_val = row[row_index]; 

                cur_bin = ctx.bin_map[cur_val];
                offset = ctx.offset_map[cur_val];

                total_weight = BIN_SIZE/256;
                total_value = total_weight * (float)cur_val;

                // work up.
                {
                    uint32_t dir_offset(BIN_SIZE - offset);
                    uint32_t this_bin(cur_bin);
                    uint32_t bin_counter(0);

                    while((this_bin < NUM_BINS) && 
                          (ctx.offset_weights[bin_counter][dir_offset] >= 0))
                    {
                        float this_weight, this_value;
                        this_weight = ctx.offset_weights[bin_counter][dir_offset] * bins[this_bin];
                        this_value = this_weight * ctx.bin_value[this_bin];

                        total_weight+= this_weight;
                        total_value+= this_value;
                        this_bin++;
                        bin_counter++;
                    }
                }
                // work down.
                {
                    uint32_t dir_offset(offset);
                    uint32_t this_bin(cur_bin);
                    uint32_t bin_counter(0);

                    while((this_bin < NUM_BINS) && 
                          (ctx.offset_weights[bin_counter][dir_offset] >= 0))
                    {
                        float this_weight, this_value;
                        this_weight = ctx.offset_weights[bin_counter][dir_offset] * bins[this_bin];
                        this_value = this_weight * ctx.bin_value[this_bin];
                        total_weight+= this_weight;
                        total_value+= this_value;
                        this_bin--;
                        bin_counter++;
                    }
                }

                if(total_weight > 0)
                {
                    value = trunc(total_value/total_weight);
                    if(value > 255)
                    {
                        value = 255;
                    }
                }
                else
                {
                    value = 0;
                }
                {
                    uint8_t val = value & 255;
                    row[row_index] = val;
                }
                row_index+= channels;
            }

            progress = (double)y/(double)height;
            progress*= (max_progress-min_progress);
            progress+= min_progress;
        }
    }
}

void tile_and_filter(const s3d_merge::Image8 *source,
                     const filter_context &ctx,
                     uint32_t tile_size,
                     uint32_t radius,
                     uint32_t channel,
                     double min_progress,
                     double max_progress,
                     s3d_merge::Image8 *dest)
{
    uint32_t effective_tile_size;
    uint32_t x, y;
    uint32_t num_tiles, tiles_processed;

    effective_tile_size = tile_size - (radius * 2);
    num_tiles = ((dest->get_width() / effective_tile_size) + 1) *
                ((dest->get_height() / effective_tile_size) + 1);
    x = y = tiles_processed = 0;

    do
    {
        uint32_t next_y;

        next_y = y + effective_tile_size;

        if(next_y > dest->get_height())
        {
            next_y = dest->get_height();
        }

        x = 0;
        do
        {
            uint32_t next_x;
            s3d_merge::IntegralHistogram *hist;

            next_x = x + effective_tile_size;

            if(next_x > dest->get_width())
            {
                next_x = dest->get_width();
            }

            hist = NULL;
            {
                uint32_t xmax, ymax;
                xmax = x + tile_size;
                ymax = y + tile_size;
                if(xmax > source->get_width())
                {
                    xmax = source->get_width();
                }

                if(ymax > source->get_height())
                {
                    ymax = source->get_height();
                }

                printf("Processing tile (%i,%i) to (%i,%i)\n", x, y, xmax, ymax);
                hist = new s3d_merge::IntegralHistogram(NUM_BINS, *source, x, y, 
                                                        xmax - x, 
                                                        ymax - y,
                                                        bin_convert,
                                                        (void *)&channel);
            }
            if(hist)
            {
                double min, max;
                min = tiles_processed;
                max = tiles_processed + 1;
                min/= num_tiles;
                max/= num_tiles;
                min*= (max_progress - min_progress);
                max*= (max_progress - min_progress);
                min+= min_progress;
                max+= min_progress;

                filter_tile(hist, ctx, radius, x, y, next_x - x, next_y - y, channel, min, max, dest);
                delete hist;
            }

            tiles_processed++;

            x = next_x;
        } while(x < dest->get_width());
        y = next_y;

    } while(y < dest->get_height());
}


s3d_merge::Image8 *
s3d_merge::BilateralFilter(uint32_t radius,
                 uint32_t threshold,
                 uint32_t tile_size,
                 bool use_linear,
                 s3d_merge::Image8 *image)
{
    Image8 *result(NULL);

    if(image)
    {
        uint32_t channels(image->get_channels());

        Image8 *source(NULL);

        // Build a source image.

        result = new Image8(*image);

        source = image->expand(radius, radius, false, false);

        if(source && result)
        {
            filter_context ctx;

            initialise_filter_context(threshold, (use_linear == 0), ctx);

            for(uint32_t i=0;i<channels;i++)
            {
                float min_progress, max_progress;
                min_progress = (double)i/(double)channels;
                max_progress = (double)(i+1)/(double)channels;

                printf("crunching channel %i\n", i);

                tile_and_filter(source, ctx, tile_size, radius, i, 
                                min_progress, max_progress, result);
            }
        }

        // clean up.
        if(source)
        {
            delete source;
        }
    }

    return result;
}


