#ifndef __SPECTRAL_FIT_H__
#define __SPECTRAL_FIT_H__

#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif

// Write out a polynomial to stdout.
// Arguments:
// "parameters" - vector of polynomial coefficients. p[0]+p[1]x+p[2]x^2 etc
// "nparameters" - size of parameters vector
// output is in a particularly unwieldy form, as it is designed to be pasted
// directly into programs or gnuplot.
void 
write_polynomial(const double *parameters, size_t nparameters);

// Evaluate a polynomial at point x.
// Arguments:
// "x"          - function argument
// "parameters" - vector of polynomial coefficients. p[0]+p[1]x+p[2]x^2 etc
// "nparameters" - size of parameters vector
// Return Value: value at point x.
double
evaluate_polynomial(double x, const double *parameters, size_t nparameters);

// Differentiate a polynomial at point x.
// Arguments:
// "x"          - function argument
// "parameters" - vector of polynomial coefficients. p[0]+p[1]x+p[2]x^2 etc
// "nparameters" - size of parameters vector
// Return Value: value at point x.
double
evaluate_differential(double x, const double *parameters, size_t nparameters);

// Generic fit function using RANSAC.
// Arguments:
// "function" - an error function, which should return rms error of some
//            subset of the data.
//            parameters are as follows:
//            "data" - void pointer to whatever it needs.
//            "items" - indices of items included in this optimisation pass
//            "nitems" - the number of items in "items"
//            "parameters" - parameters vector to optimise.
//            "nparameters" - size of parameter vector to optimise.
// "parameters"  - parameters vector to optimise.
// "nparameters" - size of parameter vector to optimise.
// "min_error"   - terminate least square step when squared error drops below this
// "convergence" - terminate when change in squared error between steps drops
//                 below this
// "max_steps"   - termination threshold for optimiser
// "max_iterations" - number of ransac iterations.
// Return Value: number of consensus points.

size_t fit_ransac(
    double (*function)(const void *, const size_t *, size_t, const double *, size_t),
    const void *data,
    size_t nitems,
    double *parameters,
    size_t nparameters,
    double min_error,
    double convergence,
    double max_steps,
    size_t nransac_items,
    double consensus_error,
    size_t iterations);
// Generic fit function.
// Arguments:
// "function" - an error function, which should return a rms error of some
//            sort.
//            parameters are as follows:
//            "data" - void pointer to whatever it needs.
//            "parameters" - parameters vector to optimise.
//            "nparameters" - size of parameter vector to optimise.
// "parameters" - parameters vector to optimise.
// "nparameters" - size of parameter vector to optimise.
// "min_error"   - terminate when squared error drops below this
// "convergence" - terminate when change in squared error between steps drops
//                 below this
// "max_steps"   - termination threshold
// Return Value: squared error, or -1 if arguments are invalid.
double
fit_parameters(double (*function)(const void *, const double *, size_t), 
               const void *data,
               double *parameters,
               size_t nparameters,
               double min_error,
               double convergence,
               size_t max_steps);

// Fit a subset of a parameter vector.
// This may be used as a shortcut for providing simple constraints.
// Arguments:
// "function" - an error function, which should return a squared error of some
//            sort.  Mean squared is good.
//            parameters are as follows:
//            "data" - void pointer to whatever it needs.
//            "parameters" - parameters vector to optimise.
//            "nparameters" - size of parameter vector to optimise.
// "parameters" - parameters vector to optimise.
// "nparameters" - size of parameter vector to optimise.
// "used"        - parameter indices to optimise
// "nused"       - number of parameter indices.
// "min_error"   - terminate when squared error drops below this
// "convergence" - terminate when change in suared error between steps drops
//                 below this
// "max_steps"   - termination threshold
// Return Value: squared error, or -1 if arguments are invalid.
double
fit_subset(double (*function)(const void *, const double *, size_t), 
           const void *data,
           double *parameters,
           size_t nparameters,
           size_t *used,
           size_t nused,
           double min_error,
           double convergence,
           size_t max_steps);

// Fit a polynomial function to a set of points.
// Arguments
// "xvalues"    - vector of x coordinates.
// "yvalues"    - vector of y coordinates.
// "nvalues"    - number of input values.
// "parameters" - parameters vector to optimise.
// "nparameters" - size of parameter vector to optimise.
// "min_error"   - terminate when squared error drops below this
// "convergence" - terminate when change in suared error between steps drops
//                 below this
// "max_steps"   - termination threshold
double
fit_polynomial_to_points(const double *xvalues,
                         const double *yvalues,
                         size_t nvalues,
                         double *parameters,
                         size_t nparameters,
                         double min_error,
                         double convergence,
                         size_t max_setps);

#ifdef __cplusplus
}
#endif

#endif

