#include "image_io.h"
#include "image_types.h"
#include "image_data.h"

#include <FreeImage.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>

#include <sys/types.h>
#include <sys/stat.h>

FREE_IMAGE_FORMAT GetFileType(const char *fn)
{
    FREE_IMAGE_FORMAT fmt = FIF_PNG;

    if(strrchr(fn, '.'))
    {
        std::string ext(strrchr(fn, '.'));

        if(
           (ext == std::string(".png")) || 
           (ext == std::string(".PNG")))
        {
            fmt = FIF_PNG;
        }
        else if(
           (ext == std::string(".tif")) || 
           (ext == std::string(".TIF")) || 
           (ext == std::string(".TIFF")) || 
           (ext == std::string(".tiff")))
        {
            fmt = FIF_TIFF;
        }
        else if(
           (ext == std::string(".ppm")) || 
           (ext == std::string(".PPM")))
        {
            fmt = FIF_PPM;
        }
    }

    return fmt;
}

// Load an image file and return an image data block.
s3d_merge::ImageData *
s3d_merge::LoadImageData(const char *fn)
{
    struct stat filedata;
    ImageData *data(NULL);

    stat(fn, &filedata);

    if(S_ISREG(filedata.st_mode))
    {
        FREE_IMAGE_FORMAT fmt = FIF_UNKNOWN;
        fmt = FreeImage_GetFileType(fn, 0);
        if((fmt != FIF_UNKNOWN) &&
           (FreeImage_FIFSupportsReading(fmt)))
        {
            FIBITMAP *bitmap = FreeImage_Load(fmt, fn, 0);

            if(bitmap)
            {
#ifdef DEBUG_IO
                fprintf(stderr, "Loaded image \"%s\" with type %i\n", 
                        fn, FreeImage_GetImageType(bitmap));
                fprintf(stderr, "Using %i colours\n",
                        FreeImage_GetColorsUsed(bitmap));
#endif
                FIMETADATA *mdhandle(NULL);
                FITAG      *tag(NULL);
                uint32_t iso(100);
                double f_number(1), 
                       exposure_time(1);
                uint32_t flags(0);

                mdhandle = FreeImage_FindFirstMetadata(FIMD_EXIF_EXIF, 
                                                       bitmap, &tag);
                if(mdhandle)
                {
                    do
                    {
                        switch(FreeImage_GetTagID(tag))
                        {
                            case EXIF_EXPOSURE_TIME:
                                {
                                    uint32_t num, denom;
                                    const uint32_t *p = (const uint32_t*)FreeImage_GetTagValue(tag);
                                    num = p[0];
                                    denom = p[1];

                                    double val = double(num);
                                    val/= double(denom);
#ifdef DEBUG_IO
                                    fprintf(stderr, 
                                            "Exposure time = %x/%x = %g seconds\n", 
                                            num, denom, val);
#endif
                                    exposure_time = val;
                                    flags |= 1;
                                }
                                break;
                            case EXIF_F_NUMBER:
                                {
                                    //uint32_t num, denom;
                                    const uint32_t *p = (const uint32_t*)FreeImage_GetTagValue(tag);
                                    double val = p[0];
                                    val/= p[1];

#ifdef DEBUG_IO
                                    fprintf(stderr, "F number = %g\n", val);
#endif
                                    f_number = val;
                                    flags |= 2;
                                }
                                break;
                            case EXIF_ISO_SPEED_RATINGS:
                                {
                                    const uint16_t *p = (const uint16_t*)FreeImage_GetTagValue(tag);

#ifdef DEBUG_IO
                                    fprintf(stderr, "ISO sensitivity = %i\n", *p);
#endif
                                    iso = *p;
                                    flags |= 4;
                                }
                                break;
                            default:
                                //printf("unhandled tag \"%s\" (0x%x)\n", 
                                //       FreeImage_GetTagKey(tag),
                                //       FreeImage_GetTagID(tag));
                                break;
                        }
                    } while(FreeImage_FindNextMetadata(mdhandle, &tag));

                    FreeImage_FindCloseMetadata(mdhandle);
                }

                {     
                    double ev(0);
                    bool has_ev(false);
                    if(flags == 7)
                    {
                        has_ev = true;
                        ev = log2((f_number * f_number) / (exposure_time * iso));
#ifdef DEBUG_IO
                        fprintf(stderr, "Exposure = %g EV\n", ev);
#endif              
                    }
                    FREE_IMAGE_TYPE image_type = FreeImage_GetImageType(bitmap);
                    printf("image_type = %i\n", image_type);

                    {
                        // Copy data to image buffer.
                        FIBITMAP *bitmap48 = FreeImage_ConvertToRGB16(bitmap);
                        if(bitmap48)
                        {
                            size_t idx(0);
                            uint32_t width = FreeImage_GetWidth(bitmap);
                            uint32_t height = FreeImage_GetHeight(bitmap);
                            Image16 *img = new Image16(width, height);
                            //FreeImage_Save(FIF_JPEG, bitmap48, "poo.jpg", 0);
                            for(uint32_t y=0;y<height;y++)
                            {
                                uint16_t *pixel;
                                pixel = (uint16_t *)FreeImage_GetScanLine(bitmap48, ((height - y) - 1));
                                for(uint32_t x=0;x<width;x++)
                                {
                                    img->get_buffer()[idx] = pixel[FI_RGBA_RED];
                                   // img->get_buffer()[idx]<<= 8;
                                    idx++;
                                    img->get_buffer()[idx] = pixel[FI_RGBA_GREEN];
                                    //img->get_buffer()[idx]<<= 8;
                                    idx++;
                                    img->get_buffer()[idx] = pixel[FI_RGBA_BLUE];
                                    //img->get_buffer()[idx]<<= 8;
                                    idx++;
                                    pixel+= 3;
                                }
                            }
                            FreeImage_Unload(bitmap48);
                            data = new ImageData(fn, img, ev);
                        }
                        else
                        {
                            fprintf(stderr, "Failed to convert bitmap %p to 32 bit.\n", bitmap);
                        }
                    }
                }

                // Clean up.
                FreeImage_Unload(bitmap);
            }

        }
        else
        {
            fprintf(stderr, "file %s is not supported\n", fn);
        }
    }
    else
    {
        fprintf(stderr, "file %s does not exist\n", fn);
    }
    return data;
}

bool
s3d_merge::WriteImageLDR(const char *fn, const s3d_merge::Image16 *img)
{
    bool written(false);

    if(fn && img)
    {
        FREE_IMAGE_FORMAT fmt = FIF_UNKNOWN;
        fmt = GetFileType(fn);

        if(fmt != FIF_UNKNOWN)
        {
            FIBITMAP *bitmap48 = FreeImage_AllocateT(FIT_RGB16, img->get_width(), img->get_height());

            uint16_t *p = img->get_buffer();

            for(size_t y=0;y<img->get_height();y++)
            {
                uint16_t *pixel;
                pixel = (uint16_t *)FreeImage_GetScanLine(bitmap48, (img->get_height() - y) - 1);
                for(size_t x=0;x<img->get_width();x++)
                {
                    uint16_t val;

                    val = *p;
                    //val>>= 8;
                    p++;
                    pixel[FI_RGBA_RED] = val;

                    val = *p;
                    //val>>= 8;
                    p++;
                    pixel[FI_RGBA_GREEN] = val;

                    val = *p;
                    //val>>= 8;
                    p++;
                    pixel[FI_RGBA_BLUE] = val;

                    pixel+= 3;
                }
            }
            FreeImage_Save(fmt, bitmap48, fn, 0);
            written = true;
            FreeImage_Unload(bitmap48);

#ifdef DEBUG_IO
            fprintf(stderr, "Saved image \"%s\"\n", fn);
#endif
        }
        else
        {
            fprintf(stderr, "unknown format for image \"%s\"\n", fn);
        }
    }
    return written;
}

bool
s3d_merge::WriteImageLDR(const char *fn, const s3d_merge::FloatImage *img)
{
    bool written(false);
    if(img)
    {
        s3d_merge::Image16 *out_image = new s3d_merge::Image16(img->get_width(),
                                                               img->get_height(),
                                                               3);
        if(out_image)
        {
            float *ip = img->get_buffer();
            uint16_t *op = out_image->get_buffer();
            size_t items = img->get_width() * img->get_height() * 3;

            for(size_t j=0;j<items;j++)
            {
                uint32_t val;
                if((*ip) <= 0.0)
                {
                    val = 0;
                }
                else if((*ip >= 1.0))
                {
                    val = 65535;
                }
                else
                {
                    val = floor((*ip) * 65536);
                }
                if(val > 65535)
                {
                    val = 65535;
                }

                *op = uint16_t(val);
                ip++;
                op++;
            }
            written = WriteImageLDR(fn, out_image);
            delete out_image;
        }
    }
    return written;
}
